import json
from copy import deepcopy
from typing import Iterator, Literal, Self

import anthropic
import backoff
import modal
import openai
from loguru import logger
from pydantic import BaseModel

from sweepai.core.entities import Message, Function
from sweepai.core.prompts import (
    system_message_prompt,
    system_message_issue_comment_prompt,
)
from sweepai.utils.chat_logger import ChatLogger
from sweepai.utils.config.server import UTILS_MODAL_INST_NAME, ANTHROPIC_API_KEY, OPENAI_DO_HAVE_32K_MODEL_ACCESS
from sweepai.utils.prompt_constructor import HumanMessagePrompt

# TODO: combine anthropic and openai

AnthropicModel = (
        Literal["claude-v1"]
        | Literal["claude-v1.3-100k"]
        | Literal["claude-instant-v1.1-100k"]
)
OpenAIModel = Literal["gpt-3.5-turbo"] | Literal["gpt-4"] | Literal["gpt-4-0613"] | Literal["gpt-3.5-turbo-16k-0613"] | Literal["gpt-4-32k"] | Literal["gpt-4-32k-0613"]

ChatModel = OpenAIModel | AnthropicModel
model_to_max_tokens = {
    "gpt-3.5-turbo": 4096,
    "gpt-4": 8192,
    "gpt-4-0613": 8192,
    "claude-v1": 9000,
    "claude-v1.3-100k": 100000,
    "claude-instant-v1.3-100k": 100000,
    "gpt-3.5-turbo-16k-0613": 16000,
    "gpt-4-32k-0613": 32000,
    "gpt-4-32k": 32000,

}
temperature = 0.1

def format_for_anthropic(messages: list[Message]) -> str:
    if len(messages) > 1:
        new_messages: list[Message] = [Message(role="system", content=messages[0].content + "\n" + messages[1].content)]
        messages = messages[2:] if len(messages) >= 3 else []
    else:
        new_messages: list[Message] = []
    for message in messages:
        new_messages.append(message)
    return "\n".join(
        f"{anthropic.HUMAN_PROMPT if message.role != 'assistant' else anthropic.AI_PROMPT} {message.content}"
        for message in new_messages
    ) + (anthropic.AI_PROMPT if new_messages[-1].role != "assistant" else "")


class ChatGPT(BaseModel):
    messages: list[Message] = [
        Message(
            role="system",
            content=system_message_prompt,
        )
    ]
    prev_message_states: list[list[Message]] = []
    model: ChatModel = "gpt-4-32k-0613" if OPENAI_DO_HAVE_32K_MODEL_ACCESS else "gpt-4-0613"
    human_message: HumanMessagePrompt | None = None
    file_change_paths = []
    chat_logger: ChatLogger | None

    @classmethod
    def from_system_message_content(
            cls, human_message: HumanMessagePrompt, is_reply: bool = False, chat_logger = None, **kwargs
    ) -> Self:
        if is_reply:
            system_message_content = system_message_issue_comment_prompt

        # Todo: This moves prompts away from unified system message prompt
        # system_message_prompt + "\n\n" + human_message.construct_prompt()
        messages = [
            Message(role="system", content=system_message_prompt, key="system")
        ]

        added_messages = human_message.construct_prompt()  # [ { role, content }, ... ]
        for msg in added_messages:
            messages.append(Message(**msg))

        return cls(
            messages=messages,
            human_message=human_message,
            chat_logger=chat_logger,
            **kwargs,
        )

    @classmethod
    def from_system_message_string(cls, prompt_string, **kwargs) -> Self:
        return cls(
            messages=[Message(role="system", content=prompt_string, key="system")],
            **kwargs,
        )

    def select_message_from_message_key(
            self, message_key: str, message_role: str = None
    ):
        if message_role:
            return [
                message
                for message in self.messages
                if message.key == message_key and message.role == message_role
            ][0]
        return [message for message in self.messages if message.key == message_key][0]

    def delete_messages_from_chat(self, key_to_delete: str):
        self.messages = [
            message for message in self.messages if key_to_delete not in (message.key or '')
        ]

    def delete_file_from_system_message(self, file_path: str):
        self.human_message.delete_file(file_path)

    def get_message_content_from_message_key(
            self, message_key: str, message_role: str = None
    ):
        return self.select_message_from_message_key(
            message_key, message_role=message_role
        ).content

    def update_message_content_from_message_key(
            self, message_key: str, new_content: str, message_role: str = None
    ):
        self.select_message_from_message_key(
            message_key, message_role=message_role
        ).content = new_content

    def chat(
            self,
            content: str,
            model: ChatModel | None = None,
            message_key: str | None = None,
            functions: list[Function] = [],
            function_name: dict | None = None,
    ):
        if self.messages[-1].function_call is None:
            self.messages.append(Message(role="user", content=content, key=message_key))
        else:
            name = self.messages[-1].function_call["name"]
            self.messages.append(Message(role="function", content=content, key=message_key, name=name))
        model = model or self.model
        is_function_call = False
        if model in [args.__args__[0] for args in OpenAIModel.__args__]:
            # might be a bug here in all of this
            if functions:
                response = self.call_openai(model=model, functions=functions, function_name=function_name)
                response, is_function_call = response
                if is_function_call:
                    self.messages.append(
                        Message(role="assistant", content=None, function_call=response, key=message_key)
                    )
                    self.prev_message_states.append(self.messages)
                    return self.messages[-1].function_call
                else:
                    self.messages.append(
                        Message(role="assistant", content=response, key=message_key)
                    )
            else:
                response = self.call_openai(model=model, functions=functions)
                self.messages.append(
                    Message(role="assistant", content=response, key=message_key)
                )
        else:
            response = self.call_anthropic(model=model)
            self.messages.append(
                Message(role="assistant", content=response, key=message_key)
            )
        self.prev_message_states.append(self.messages)
        return self.messages[-1].content

    def call_openai(
            self,
            model: ChatModel | None = None,
            functions: list[Function] = [],
            function_name: dict | None = None,
    ):
        if self.chat_logger:
            tickets_allocated = 60 if self.chat_logger.is_paying_user() else 3
            tickets_count = self.chat_logger.get_ticket_count()
            if tickets_count < tickets_allocated:
                model = model or self.model
                logger.warning(f"{tickets_count} tickets found in MongoDB, using {model}")
            else:
                model = "gpt-3.5-turbo-16k-0613"
        else:
            model = "gpt-3.5-turbo-16k-0613"

        count_tokens = modal.Function.lookup(UTILS_MODAL_INST_NAME, "Tiktoken.count")
        messages_length = sum(
            [count_tokens.call(message.content or "") for message in self.messages]
        )
        max_tokens = model_to_max_tokens[model] - int(messages_length) - 400  # this is for the function tokens
        # TODO: Add a check to see if the message is too long
        logger.info("file_change_paths" + str(self.file_change_paths))
        if len(self.file_change_paths) > 0:
            self.file_change_paths.remove(self.file_change_paths[0])
        if max_tokens < 0:
            if len(self.file_change_paths) > 0:
                pass
            else:
                raise ValueError(f"Message is too long, max tokens is {max_tokens}")
        messages_raw = "\n".join([(message.content or "") for message in self.messages])
        logger.info(f"Input to call openai:\n{messages_raw}")

        gpt_4_buffer = 800
        if int(messages_length) + gpt_4_buffer < 6000 and model == "gpt-4-32k-0613":
            model = "gpt-4-0613"
            max_tokens = model_to_max_tokens[model] - int(
                messages_length) - gpt_4_buffer  # this is for the function tokens
        if "gpt-4" in model:
            max_tokens = min(max_tokens, 5000)
        logger.info(f"Using the model {model}, with {max_tokens} tokens remaining")
        global retry_counter
        retry_counter = 0
        if functions:
            @backoff.on_exception(
                backoff.expo,
                Exception,
                max_tries=12,
                jitter=backoff.random_jitter,

            )
            def fetch():
                global retry_counter
                retry_counter += 1
                token_sub = retry_counter * 200
                try:
                    output = None
                    if function_name:
                        output = (
                            openai.ChatCompletion.create(
                                model=model,
                                messages=self.messages_dicts,
                                max_tokens=max_tokens - token_sub,
                                temperature=temperature,
                                functions=[json.loads(function.json()) for function in functions],
                                function_call=function_name,
                            )
                            .choices[0].message
                        )
                    else:
                        output = (
                            openai.ChatCompletion.create(
                                model=model,
                                messages=self.messages_dicts,
                                max_tokens=max_tokens - token_sub,
                                temperature=temperature,
                                functions=[json.loads(function.json()) for function in functions],
                            )
                            .choices[0].message
                        )
                    if self.chat_logger is not None: self.chat_logger.add_chat({
                        'model': model,
                        'messages': self.messages_dicts,
                        'max_tokens': max_tokens - token_sub,
                        'temperature': temperature,
                        'functions': [json.loads(function.json()) for function in functions],
                        'function_call': function_name,
                        'output': output,
                    })
                    return output
                except Exception as e:
                    logger.warning(e)
                    raise e

            result = fetch()
            if "function_call" in result:
                result = dict(result["function_call"]), True
            else:
                result = result["content"], False
            logger.info(f"Output to call openai:\n{result}")
            return result

        else:
            @backoff.on_exception(
                backoff.expo,
                Exception,
                max_tries=12,
                jitter=backoff.random_jitter,
            )
            def fetch():
                global retry_counter
                retry_counter += 1
                token_sub = retry_counter * 200
                try:
                    output = openai.ChatCompletion.create(
                        model=model,
                        messages=self.messages_dicts,
                        max_tokens=max_tokens - token_sub,
                        temperature=temperature,
                    ) \
                        .choices[0] \
                        .message["content"]
                    if self.chat_logger is not None: self.chat_logger.add_chat({
                        'model': model,
                        'messages': self.messages_dicts,
                        'max_tokens': max_tokens - token_sub,
                        'temperature': temperature,
                        'output': output
                    })
                    return output
                except Exception as e:
                    logger.warning(e)
                    raise e

            result = fetch()
            logger.info(f"Output to call openai:\n{result}")
            return result

    def call_anthropic(self, model: ChatModel | None = None) -> str:
        if model is None:
            model = self.model
        count_tokens = modal.Function.lookup(UTILS_MODAL_INST_NAME, "Tiktoken.count")
        messages_length = sum(
            [int(count_tokens.call(message.content) * 1.1) for message in self.messages]
        )
        max_tokens = model_to_max_tokens[model] - int(messages_length) - 1000
        logger.info(f"Number of tokens: {max_tokens}")
        messages_raw = format_for_anthropic(self.messages)
        logger.info(f"Input to call anthropic:\n{messages_raw}")

        assert ANTHROPIC_API_KEY is not None
        client = anthropic.Client(api_key=ANTHROPIC_API_KEY)

        @backoff.on_exception(
            backoff.expo,
            Exception,
            max_tries=12,
            jitter=backoff.random_jitter,
        )
        def fetch() -> tuple[str, str]:
            logger.warning(f"Calling anthropic...")
            results = client.completion(
                prompt=messages_raw,
                stop_sequences=[anthropic.HUMAN_PROMPT],
                model=model,
                max_tokens_to_sample=max_tokens,
                disable_checks=True,
                temperature=temperature,
            )
            return results["completion"], results["stop_reason"]

        result, stop_reason = fetch()
        logger.warning(f"Stop reasons: {stop_reason}")
        if stop_reason == "max_tokens":
            logger.warning("Hit max tokens, running for more tokens.")
            _self = deepcopy(self)
            _self.messages.append(Message(role="assistant", content=result, key=""))
            extension = _self.call_anthropic(model=model)
            print(len(result), len(extension), len(result + extension))
            return result + extension
        logger.info(f"Output to call anthropic:\n{result}")
        return result

    def chat_stream(
            self,
            content: str,
            model: ChatModel | None = None,
            message_key: str | None = None,
            functions: list[Function] = [],
            function_call: dict | None = None,
    ) -> Iterator[dict]:
        if self.messages[-1].function_call is None:
            self.messages.append(Message(role="user", content=content, key=message_key))
        else:
            name = self.messages[-1].function_call["name"]
            self.messages.append(Message(role="function", content=content, key=message_key, name=name))
        model = model or self.model
        is_function_call = False
        # might be a bug here in all of this
        # return self.stream_openai(model=model, functions=functions, function_name=function_name)
        return self.stream_openai(model=model, functions=functions, function_call=function_call)

    def stream_openai(
            self,
            model: ChatModel | None = None,
            functions: list[Function] = [],
            function_call: dict | None = None,
    ) -> Iterator[dict]:
        model = model or self.model
        count_tokens = modal.Function.lookup(UTILS_MODAL_INST_NAME, "Tiktoken.count")
        messages_length = sum(
            [count_tokens.call(message.content or "") for message in self.messages]
        )
        max_tokens = model_to_max_tokens[model] - int(messages_length) - 400  # this is for the function tokens
        # TODO: Add a check to see if the message is too long
        logger.info("file_change_paths" + str(self.file_change_paths))
        if len(self.file_change_paths) > 0:
            self.file_change_paths.remove(self.file_change_paths[0])
        if max_tokens < 0:
            if len(self.file_change_paths) > 0:
                pass
            else:
                raise ValueError(f"Message is too long, max tokens is {max_tokens}")
        messages_raw = "\n".join([(message.content or "") for message in self.messages])
        logger.info(f"Input to call openai:\n{messages_raw}")

        gpt_4_buffer = 800
        if int(messages_length) + gpt_4_buffer < 6000 and model == "gpt-4-32k-0613":
            model = "gpt-4-0613"
            max_tokens = model_to_max_tokens[model] - int(
                messages_length) - gpt_4_buffer  # this is for the function tokens

        logger.info(f"Using the model {model}, with {max_tokens} tokens remaining")

        def generator() -> Iterator[str]:
            stream = openai.ChatCompletion.create(
                model=model,
                messages=self.messages_dicts,
                temperature=temperature,
                functions=[json.loads(function.json()) for function in functions],
                function_call=function_call or "auto",
                stream=True
            ) if functions else openai.ChatCompletion.create(
                model=model,
                messages=self.messages_dicts,
                temperature=temperature,
                stream=True
            )
            for data in stream:
                chunk = data.choices[0].delta
                yield chunk

        return generator()

    @property
    def messages_dicts(self):
        # Remove the key from the message object before sending to OpenAI
        cleaned_messages = [
            message.to_openai()
            for message in self.messages
        ]
        return cleaned_messages

    def undo(self):
        if len(self.prev_message_states) > 0:
            self.messages = self.prev_message_states.pop()
        return self.messages



import io
import os
import zipfile

import openai
import requests
from loguru import logger

from sweepai.core.gha_extraction import GHAExtractor
from sweepai.events import CheckRunCompleted
from sweepai.utils.config.client import SweepConfig, get_gha_enabled
from sweepai.utils.github_utils import get_github_client, get_token

openai.api_key = os.environ.get("OPENAI_API_KEY")


def download_logs(repo_full_name: str, run_id: int, installation_id: int):
    headers = {
        "Accept": "application/vnd.github+json",
        "Authorization": f"Bearer {get_token(installation_id)}",
        "X-GitHub-Api-Version": "2022-11-28"
    }
    response = requests.get(f"https://api.github.com/repos/{repo_full_name}/actions/runs/{run_id}/logs",
                            headers=headers)

    logs_str = ""
    if response.status_code == 200:
        zip_file = zipfile.ZipFile(io.BytesIO(response.content))
        for file in zip_file.namelist():
            if "/" not in file:
                with zip_file.open(file) as f:
                    logs_str += f.read().decode("utf-8")
    else:
        logger.warning(f"Failed to download logs for run id: {run_id}")
    return logs_str


def clean_logs(logs_str: str):
    log_list = logs_str.split("\n")
    truncated_logs = [log[log.find(" ") + 1:] for log in log_list]
    patterns = [
        # for docker
        "Already exists",
        "Pulling fs layer",
        "Waiting",
        "Download complete",
        "Verifying Checksum",
        "Pull complete",
        # For github
        "remote: Counting objects",
        "remote: Compressing objects:",
        "Receiving objects:",
        "Resolving deltas:"
    ]
    return "\n".join([log.strip() for log in truncated_logs if not any(pattern in log for pattern in patterns)])


def on_check_suite(request: CheckRunCompleted):
    g = get_github_client(request.installation.id)
    repo = g.get_repo(request.repository.full_name)
    if not get_gha_enabled(repo):
        return None
    pr = repo.get_pull(request.check_run.pull_requests[0].number)
    num_pr_commits = len(list(pr.get_commits()))
    if num_pr_commits > 20:
        return None
    logger.info(f"Running github action for PR with {num_pr_commits} commits")
    logs = download_logs(
        request.repository.full_name,
        request.check_run.run_id,
        request.installation.id
    )
    if not logs:
        return None
    logs = clean_logs(logs)
    extractor = GHAExtractor()
    logger.info(f"Extracting logs from {request.repository.full_name}, logs: {logs}")
    logs = extractor.gha_extract(logs)
    return logs



import modal
from fastapi import HTTPException, Request
from loguru import logger
from pydantic import ValidationError

from sweepai.events import (
    CheckRunCompleted,
    CommentCreatedRequest,
    InstallationCreatedRequest,
    IssueCommentRequest,
    IssueRequest,
    PRRequest,
    ReposAddedRequest,
)
from sweepai.handlers.create_pr import create_pr  # type: ignore
from sweepai.handlers.on_check_suite import on_check_suite  # type: ignore
from sweepai.handlers.on_comment import on_comment
from sweepai.handlers.on_ticket import on_ticket
from sweepai.utils.config.server import DB_MODAL_INST_NAME, API_MODAL_INST_NAME, GITHUB_BOT_USERNAME, \
    GITHUB_LABEL_NAME, GITHUB_LABEL_COLOR, GITHUB_LABEL_DESCRIPTION, BOT_TOKEN_NAME
from sweepai.utils.event_logger import posthog
from sweepai.utils.github_utils import get_github_client, index_full_repository

stub = modal.Stub(API_MODAL_INST_NAME)
image = (
    modal.Image.debian_slim()
    .apt_install("git")
    .pip_install(
        "openai",
        "anthropic",
        "PyGithub",
        "loguru",
        "docarray",
        "backoff",
        "tiktoken",
        "highlight-io",
        "GitPython",
        "posthog",
        "tqdm",
        "pyyaml",
        "pymongo",
        "tabulate"
    )
)
secrets = [
    modal.Secret.from_name(BOT_TOKEN_NAME),
    modal.Secret.from_name("github"),
    modal.Secret.from_name("openai-secret"),
    modal.Secret.from_name("anthropic"),
    modal.Secret.from_name("posthog"),
    modal.Secret.from_name("highlight"),
    modal.Secret.from_name("mongodb"),
    modal.Secret.from_name("discord")
]

FUNCTION_SETTINGS = {
    "image": image,
    "secrets": secrets,
    "timeout": 30 * 60,
}

handle_ticket = stub.function(**FUNCTION_SETTINGS)(on_ticket)
handle_comment = stub.function(**FUNCTION_SETTINGS)(on_comment)
handle_pr = stub.function(**FUNCTION_SETTINGS)(create_pr)
update_index = modal.Function.lookup(DB_MODAL_INST_NAME, "update_index")
handle_check_suite = stub.function(**FUNCTION_SETTINGS)(on_check_suite)


@stub.function(**FUNCTION_SETTINGS)
@modal.web_endpoint(method="POST")
async def webhook(raw_request: Request):
    """Handle a webhook request from GitHub."""
    try:
        request_dict = await raw_request.json()
        logger.info(f"Received request: {request_dict.keys()}")
        event = raw_request.headers.get("X-GitHub-Event")
        assert event is not None
        match event, request_dict.get("action", None):
            case "issues", "opened":
                request = IssueRequest(**request_dict)
                issue_title_lower = request.issue.title.lower()
                if issue_title_lower.startswith("sweep") or "sweep:" in issue_title_lower:
                    g = get_github_client(request.installation.id)
                    repo = g.get_repo(request.repository.full_name)

                    labels = repo.get_labels()
                    label_names = [label.name for label in labels]

                    if GITHUB_LABEL_NAME not in label_names:
                        repo.create_label(
                            name=GITHUB_LABEL_NAME,
                            color=GITHUB_LABEL_COLOR,
                            description=GITHUB_LABEL_DESCRIPTION,
                        )
                    # TODO(sweep): figure out why this is breaking
                    # else:
                    #     label = repo.get_label(LABEL_NAME)
                    #     label.edit(
                    #         name=LABEL_NAME,
                    #         color=LABEL_COLOR, 
                    #         description=LABEL_DESCRIPTION
                    #     )

                    current_issue = repo.get_issue(number=request.issue.number)
                    current_issue.add_to_labels(GITHUB_LABEL_NAME)
            case "issues", "labeled":
                request = IssueRequest(**request_dict)
                if 'label' in request_dict and str.lower(request_dict['label']['name']) == GITHUB_LABEL_NAME:
                    request.issue.body = request.issue.body or ""
                    request.repository.description = (
                            request.repository.description or ""
                    )
                    # Update before we handle the ticket to make sure index is up to date
                    # other ways suboptimal
                    handle_ticket.spawn(
                        request.issue.title,
                        request.issue.body,
                        request.issue.number,
                        request.issue.html_url,
                        request.issue.user.login,
                        request.repository.full_name,
                        request.repository.description,
                        request.installation.id,
                        None
                    )
            case "issue_comment", "created":
                request = IssueCommentRequest(**request_dict)
                # if replying to an issue with sweep label
                if request.issue is not None \
                        and GITHUB_LABEL_NAME in [label.name.lower() for label in request.issue.labels] \
                        and request.comment.user.type == "User":
                    request.issue.body = request.issue.body or ""
                    request.repository.description = (
                            request.repository.description or ""
                    )

                    if not request.comment.body.lower().startswith(GITHUB_LABEL_NAME):
                        return {"success": True, "reason": "Comment does not start with 'Sweep', passing"}

                    # Update before we handle the ticket to make sure index is up to date
                    # other ways suboptimal
                    handle_ticket.spawn(
                        request.issue.title,
                        request.issue.body,
                        request.issue.number,
                        request.issue.html_url,
                        request.issue.user.login,
                        request.repository.full_name,
                        request.repository.description,
                        request.installation.id,
                        request.comment.id
                    )
                elif request.issue.pull_request and request.issue.user.login == GITHUB_BOT_USERNAME and request.comment.user.type == "User":  # TODO(sweep): set a limit
                    logger.info(f"Handling comment on PR: {request.issue.pull_request}")
                    handle_comment.spawn(
                        repo_full_name=request.repository.full_name,
                        repo_description=request.repository.description,
                        comment=request.comment.body,
                        pr_path=None,
                        pr_line_position=None,
                        username=request.comment.user.login,
                        installation_id=request.installation.id,
                        pr_number=request.issue.number,
                    )
            case "pull_request_review_comment", "created":
                request = CommentCreatedRequest(**request_dict)
                if "sweep/" in request.pull_request.head.ref.lower():
                    handle_comment.spawn(
                        repo_full_name=request.repository.full_name,
                        repo_description=request.repository.description,
                        comment=request.comment.body,
                        pr_path=request.comment.path,
                        pr_line_position=request.comment.original_line,
                        username=request.comment.user.login,
                        installation_id=request.installation.id,
                        pr_number=request.pull_request.number,
                    )
                # Todo: update index on comments
            case "pull_request_review", "submitted":
                # request = ReviewSubmittedRequest(**request_dict)
                pass
            case "check_run", "completed":
                request = CheckRunCompleted(**request_dict)
                # handle_check_suite
                logs = None
                # Must be Sweep firing the PR and it must fail
                if request.sender.login == GITHUB_BOT_USERNAME and request.check_run.conclusion == "failure":
                    logs = handle_check_suite.call(request)
                if len(request.check_run.pull_requests) > 0 and logs:
                    handle_comment.spawn(
                        repo_full_name=request.repository.full_name,
                        repo_description=request.repository.description,
                        comment="Sweep: " + logs,
                        pr_path=None,
                        pr_line_position=None,
                        username=request.sender.login,
                        installation_id=request.installation.id,
                        pr_number=request.check_run.pull_requests[0].number,
                    )
            case "installation_repositories", "added":
                repos_added_request = ReposAddedRequest(**request_dict)
                metadata = {
                    "installation_id": repos_added_request.installation.id,
                    "repositories": [
                        repo.full_name
                        for repo in repos_added_request.repositories_added
                    ],
                }
                posthog.capture("installation_repositories", "started", properties={
                    **metadata
                })
                for repo in repos_added_request.repositories_added:
                    organization, repo_name = repo.full_name.split("/")
                    posthog.capture(
                        organization,
                        "installed_repository",
                        properties={
                            "repo_name": repo_name,
                            "organization": organization,
                            "repo_full_name": repo.full_name
                        }
                    )
                    index_full_repository(
                        repo.full_name,
                        installation_id=repos_added_request.installation.id,
                    )
            case "installation", "created":
                repos_added_request = InstallationCreatedRequest(**request_dict)
                for repo in repos_added_request.repositories:
                    index_full_repository(
                        repo.full_name,
                        installation_id=repos_added_request.installation.id,
                    )
            case ("pull_request", "closed"):
                pr_request = PRRequest(**request_dict)
                organization, repo_name = pr_request.repository.full_name.split("/")
                commit_author = pr_request.pull_request.user.login
                merged_by = pr_request.pull_request.merged_by.login
                if GITHUB_BOT_USERNAME == commit_author:
                    posthog.capture(
                        merged_by,
                        "merged_sweep_pr",
                        properties={
                            "repo_name": repo_name,
                            "organization": organization,
                            "repo_full_name": pr_request.repository.full_name,
                            "username": merged_by
                        })
                update_index.spawn(
                    request_dict["repository"]["full_name"],
                    installation_id=request_dict["installation"]["id"],
                )
            case ("push", None):
                if event != "pull_request" or request_dict["base"]["merged"] == True:
                    update_index.spawn(
                        request_dict["repository"]["full_name"],
                        installation_id=request_dict["installation"]["id"],
                    )
            case "ping", None:
                return {"message": "pong"}
            case _:
                logger.info(
                    f"Unhandled event: {event} {request_dict.get('action', None)}"
                )
    except ValidationError as e:
        logger.warning(f"Failed to parse request: {e}")
        raise HTTPException(status_code=422, detail="Failed to parse request")
    return {"success": True}



import glob
import json
import os
import re
import shutil
import time

import modal
from deeplake.core.vectorstore.deeplake_vectorstore import DeepLakeVectorStore
from git.repo import Repo
from github import Github
from loguru import logger
from modal import method
from redis import Redis
from tqdm import tqdm

from sweepai.core.entities import Snippet
from sweepai.utils.event_logger import posthog
from sweepai.utils.hash import hash_sha256
from sweepai.utils.scorer import compute_score, convert_to_percentiles
from ..utils.config.client import SweepConfig
from ..utils.config.server import ENV, DB_MODAL_INST_NAME, UTILS_MODAL_INST_NAME, REDIS_URL, BOT_TOKEN_NAME
from ..utils.github_utils import get_token

# TODO: Lots of cleanups can be done here with these constants
stub = modal.Stub(DB_MODAL_INST_NAME)
chunker = modal.Function.lookup(UTILS_MODAL_INST_NAME, "Chunking.chunk")
model_volume = modal.SharedVolume().persist(f"{ENV}-storage")
MODEL_DIR = "/root/cache/model"
DEEPLAKE_DIR = "/root/cache/"
DISKCACHE_DIR = "/root/cache/diskcache/"
DEEPLAKE_FOLDER = "deeplake/"
BATCH_SIZE = 256
SENTENCE_TRANSFORMERS_MODEL = "sentence-transformers/all-MiniLM-L12-v2"
timeout = 60 * 30  # 30 minutes
CACHE_VERSION = "v1.0.0"
MAX_FILES = 3000

image = (
    modal.Image.debian_slim()
    .apt_install("git")
    .pip_install("deeplake==3.6.3", "sentence-transformers")
    .pip_install("openai", "PyGithub", "loguru", "docarray", "GitPython", "tqdm", "highlight-io", "anthropic",
                 "posthog", "redis", "pyyaml")
)
secrets = [
    modal.Secret.from_name(BOT_TOKEN_NAME),
    modal.Secret.from_name("github"),
    modal.Secret.from_name("openai-secret"),
    modal.Secret.from_name("huggingface"),
    modal.Secret.from_name("chroma-endpoint"),
    modal.Secret.from_name("posthog"),
    modal.Secret.from_name("highlight"),
    modal.Secret.from_name("redis_url"),
    modal.Secret.from_dict({"TRANSFORMERS_CACHE": MODEL_DIR}),
]


def init_deeplake_vs(repo_name):
    deeplake_repo_path = f"mem://{DEEPLAKE_FOLDER}{repo_name}"
    deeplake_vector_store = DeepLakeVectorStore(path=deeplake_repo_path)
    return deeplake_vector_store


def parse_collection_name(name: str) -> str:
    # Replace any non-alphanumeric characters with hyphens
    name = re.sub(r"[^\w-]", "--", name)
    # Ensure the name is between 3 and 63 characters and starts/ends with alphanumeric
    name = re.sub(r"^(-*\w{0,61}\w)-*$", r"\1", name[:63].ljust(3, "x"))
    return name


@stub.cls(
    image=image,
    secrets=secrets,
    shared_volumes={MODEL_DIR: model_volume},
    keep_warm=1 if ENV == "prod" else 0,
    gpu="T4",
    retries=modal.Retries(
        max_retries=5, backoff_coefficient=2, initial_delay=5),
)
class Embedding:
    def __enter__(self):
        from sentence_transformers import SentenceTransformer

        self.model = SentenceTransformer(
            SENTENCE_TRANSFORMERS_MODEL, cache_folder=MODEL_DIR
        )

    @method()
    def compute(self, texts: list[str]):
        return self.model.encode(texts, batch_size=BATCH_SIZE).tolist()

    @method()
    def ping(self):
        return "pong"


class ModalEmbeddingFunction():
    def __init__(self):
        pass

    def __call__(self, texts):
        return Embedding.compute.call(texts)


embedding_function = ModalEmbeddingFunction()


def get_deeplake_vs_from_repo(
        repo_name: str,
        installation_id: int,
        branch_name: str | None = None,
        sweep_config: SweepConfig = SweepConfig(),
):
    token = get_token(installation_id)
    g = Github(token)
    repo = g.get_repo(repo_name)
    commits = repo.get_commits()
    commit_hash = commits[0].sha

    cache_success = False
    cache_inst = None

    if REDIS_URL is not None:
        try:
            # todo: initialize once
            cache_inst = Redis.from_url(REDIS_URL)
            logger.info(f"Successfully connected to redis cache")
            cache_success = True
        except:
            cache_success = False
            logger.error(f"Failed to connect to redis cache")
    else:
        logger.warning(f"REDIS_URL is None, skipping cache")

    if cache_inst and cache_success:

        try:
            github_cache_key = f"github-{commit_hash}{CACHE_VERSION}"
            cache_hit = cache_inst.get(github_cache_key)
            if cache_hit:
                deeplake_items = json.loads(cache_hit)
                logger.info(f"Cache hit for {repo_name}")
            else:
                deeplake_items = None
                logger.info(f"Cache miss for {repo_name}")

            if deeplake_items:
                deeplake_vs = init_deeplake_vs(repo_name)
                deeplake_vs.add(
                    text=deeplake_items['ids'],
                    embedding=deeplake_items['embeddings'],
                    metadata=deeplake_items['metadatas']
                )
                logger.info(f"Returning deeplake vs for {repo_name}")
                return deeplake_vs
            else:
                logger.info(f"Cache for {repo_name} is empty")
        except:
            logger.info(f"Failed to get cache for {repo_name}")
    logger.info(f"Downloading repository and indexing for {repo_name}...")
    start = time.time()
    logger.info("Recursively getting list of files...")

    repo_url = f"https://x-access-token:{token}@github.com/{repo_name}.git"
    shutil.rmtree("repo", ignore_errors=True)

    branch_name = SweepConfig.get_branch(repo)

    git_repo = Repo.clone_from(repo_url, "repo")
    git_repo.git.checkout(branch_name)

    file_list = glob.iglob("repo/**", recursive=True)
    file_list = [
        file
        for file in tqdm(file_list)
        if os.path.isfile(file)
           and all(not file.endswith(ext) for ext in sweep_config.exclude_exts)
           and all(not file[len("repo/"):].startswith(dir_name) for dir_name in sweep_config.exclude_dirs)
    ]

    file_paths = []
    file_contents = []
    scores = []

    for file in tqdm(file_list):
        with open(file, "rb") as f:
            is_binary = False
            for block in iter(lambda: f.read(1024), b''):
                if b'\0' in block:
                    is_binary = True
                    break
            if is_binary:
                logger.debug("Skipping binary file...")
                continue

        with open(file, "rb") as f:
            if len(f.read()) > sweep_config.max_file_limit:
                logger.debug("Skipping large file...")
                continue

        with open(file, "r") as f:
            # Can parallelize this
            try:
                contents = f.read()
                contents = file + contents
            except UnicodeDecodeError as e:
                logger.warning(f"Received warning {e}, skipping...")
                continue
            file_path = file[len("repo/"):]
            file_paths.append(file_path)
            file_contents.append(contents)
            if len(file_list) > MAX_FILES:
                scores.append(1)
                continue
            try:
                cache_key = f"{repo_name}-{file_path}-{CACHE_VERSION}"
                if cache_inst and cache_success:
                    cached_value = cache_inst.get(cache_key)
                    if cached_value:
                        score = json.loads(cached_value)
                        scores.append(score)
                        continue
                commits = list(repo.get_commits(
                    path=file_path, sha=branch_name))
                score = compute_score(contents, commits)
                if cache_inst and cache_success:
                    cache_inst.set(cache_key, json.dumps(score), ex=60 * 60 * 2)
                scores.append(score)
            except Exception as e:
                logger.warning(f"Received warning during scoring {e}, skipping...")
                scores.append(1)
                continue
    scores = convert_to_percentiles(scores)

    chunked_results = chunker.map(file_contents, file_paths, scores, kwargs={
        "additional_metadata": {"repo_name": repo_name, "branch_name": branch_name}
    })

    documents, metadatas, ids = zip(*chunked_results)
    documents = [item for sublist in documents for item in sublist]
    metadatas = [item for sublist in metadatas for item in sublist]
    ids = [item for sublist in ids for item in sublist]

    logger.info(f"Used {len(file_paths)} files...")

    shutil.rmtree("repo", ignore_errors=True)
    logger.info(f"Getting list of all files took {time.time() - start}")
    logger.info(
        f"Received {len(documents)} documents from repository {repo_name}")
    collection_name = parse_collection_name(repo_name)
    return compute_deeplake_vs(collection_name, documents, cache_success, cache_inst, ids, metadatas, commit_hash)


def compute_deeplake_vs(collection_name,
                        documents,
                        cache_success,
                        cache_inst,
                        ids,
                        metadatas,
                        sha):
    deeplake_vs = init_deeplake_vs(collection_name)
    if len(documents) > 0:
        logger.info("Computing embeddings...")
        # Check cache here for all documents
        embeddings = [None] * len(documents)
        if cache_inst and cache_success:
            cache_keys = [hash_sha256(
                doc) + SENTENCE_TRANSFORMERS_MODEL + CACHE_VERSION for doc in documents]
            cache_values = cache_inst.mget(cache_keys)
            for idx, value in enumerate(cache_values):
                if value is not None:
                    embeddings[idx] = json.loads(value)
        logger.info(
            f"Found {len([x for x in embeddings if x is not None])} embeddings in cache")
        indices_to_compute = [idx for idx,
        x in enumerate(embeddings) if x is None]
        documents_to_compute = [documents[idx] for idx in indices_to_compute]

        computed_embeddings = embedding_function(documents_to_compute)

        for idx, embedding in zip(indices_to_compute, computed_embeddings):
            embeddings[idx] = embedding
        deeplake_vs.add(
            text=ids,
            embedding=embeddings,
            metadata=metadatas
        )
        if cache_inst and cache_success:
            cache_inst.set(f"github-{sha}{CACHE_VERSION}", json.dumps(
                {"metadatas": metadatas, "ids": ids, "embeddings": embeddings}))
        if cache_inst and cache_success and len(documents_to_compute) > 0:
            logger.info(
                f"Updating cache with {len(computed_embeddings)} embeddings")
            cache_keys = [hash_sha256(
                doc) + SENTENCE_TRANSFORMERS_MODEL + CACHE_VERSION for doc in documents_to_compute]
            cache_inst.mset({key: json.dumps(value)
                             for key, value in zip(cache_keys, computed_embeddings)})
        return deeplake_vs
    else:
        logger.error("No documents found in repository")
        return deeplake_vs


@stub.function(image=image, secrets=secrets, shared_volumes={DISKCACHE_DIR: model_volume}, timeout=timeout)
def update_index(
        repo_name,
        installation_id: int,
        sweep_config: SweepConfig = SweepConfig(),
) -> int:
    get_deeplake_vs_from_repo(repo_name, installation_id, branch_name=None, sweep_config=sweep_config)
    # todo: ?
    return 0


@stub.function(image=image, secrets=secrets, shared_volumes={DEEPLAKE_DIR: model_volume}, timeout=timeout, keep_warm=1)
def get_relevant_snippets(
        repo_name: str,
        query: str,
        n_results: int,
        installation_id: int,
        username: str | None = None,
        sweep_config: SweepConfig = SweepConfig(),
):
    deeplake_vs = get_deeplake_vs_from_repo(
        repo_name=repo_name, installation_id=installation_id, sweep_config=sweep_config
    )
    results = {"metadata": [], "text": []}
    for n_result in range(n_results, 0, -1):
        try:
            query_embedding = embedding_function([query])[0]
            results = deeplake_vs.search(embedding=query_embedding, k=n_result)
            break
        except Exception:
            pass
    if len(results["text"]) == 0:
        if username is None:
            username = "anonymous"
        posthog.capture(
            username,
            "failed",
            {
                "reason": "Results query was empty",
                "repo_name": repo_name,
                "installation_id": installation_id,
                "query": query,
                "n_results": n_results
            },
        )
    metadatas = results["metadata"]
    code_scores = [metadata["score"] for metadata in metadatas]
    vector_scores = results["score"]
    combined_scores = [code_score + vector_score for code_score,
    vector_score in zip(code_scores, vector_scores)]
    # Sort by combined scores
    # Combine the three lists into a single list of tuples
    combined_list = list(zip(combined_scores, metadatas))

    # Sort the combined list based on the combined scores
    sorted_list = sorted(combined_list, key=lambda x: x[0], reverse=True)

    # Extract the sorted metadatas and relevant_paths
    sorted_metadatas = [metadata for _, metadata in sorted_list]
    relevant_paths = [metadata["file_path"] for metadata in sorted_metadatas]
    logger.info("Relevant paths: {}".format(relevant_paths))
    return [
        Snippet(
            content="",
            start=metadata["start"],
            end=metadata["end"],
            file_path=file_path
        ) for metadata, file_path in zip(sorted_metadatas, relevant_paths)
    ]



import os

PREFIX = "prod"
ENV = PREFIX

DB_MODAL_INST_NAME = PREFIX + "-db"
API_MODAL_INST_NAME = PREFIX + "-api"
UTILS_MODAL_INST_NAME = PREFIX + "-utils"
SLACK_MODAL_INST_NAME = PREFIX + "-slack"

# deprecated: old logic transfer so upstream can use this; just create an empty modal secret for this
if PREFIX == "prod":
    BOT_TOKEN_NAME = "bot-token"
else:
    BOT_TOKEN_NAME = PREFIX + "-bot-token"

# goes under Modal 'discord' secret name (optional, can leave env var blank)
DISCORD_WEBHOOK_URL = os.environ.get('DISCORD_WEBHOOK_URL')

# goes under Modal 'github' secret name
GITHUB_BOT_TOKEN = os.environ.get('GITHUB_BOT_TOKEN')
# deprecated: old logic transfer so upstream can use this
if not GITHUB_BOT_TOKEN:
    GITHUB_BOT_TOKEN = os.environ.get('GITHUB_TOKEN')

GITHUB_APP_ID = os.environ.get('GITHUB_APP_ID')
# deprecated: old logic transfer so upstream can use this
if not GITHUB_APP_ID:
    if PREFIX == "prod":
        GITHUB_APP_ID = "307814"
    elif PREFIX == "dev":
        GITHUB_APP_ID = "324098"
    elif PREFIX == "dev2":
        GITHUB_APP_ID = "327588"
GITHUB_BOT_USERNAME = os.environ.get('GITHUB_BOT_USERNAME')
# deprecated: old logic transfer so upstream can use this
if not GITHUB_BOT_USERNAME:
    if PREFIX == "prod":
        GITHUB_BOT_USERNAME = "sweep-ai[bot]"
    elif PREFIX == "dev":
        GITHUB_BOT_USERNAME = "sweep-nightly[bot]"
    elif PREFIX == "dev2":
        GITHUB_BOT_USERNAME = "sweep-canary[bot]"

GITHUB_LABEL_NAME = os.environ.get('GITHUB_LABEL_NAME', 'sweep')
GITHUB_LABEL_COLOR = os.environ.get('GITHUB_LABEL_COLOR', '9400D3')
GITHUB_LABEL_DESCRIPTION = os.environ.get('GITHUB_LABEL_DESCRIPTION', 'Sweep your software chores')
GITHUB_APP_PEM = os.environ.get('GITHUB_APP_PEM')
GITHUB_CONFIG_BRANCH = os.environ.get('GITHUB_CONFIG_BRANCH', 'sweep/add-sweep-config')
GITHUB_DEFAULT_CONFIG = os.environ.get('GITHUB_DEFAULT_CONFIG', """# Sweep AI turns bug fixes & feature requests into code changes (https://sweep.dev)
# For details on our config file, check out our docs at https://docs.sweep.dev
# Reference: https://github.com/sweepai/sweep/blob/main/.github/sweep.yaml.

# If you use this be sure to frequently sync your default branch(main, master) to dev.
branch: '{branch}'
""")

# goes under Modal 'openai-secret' secret name
OPENAI_API_KEY = os.environ.get('OPENAI_API_KEY')
OPENAI_DO_HAVE_32K_MODEL_ACCESS = os.environ.get('OPENAI_DO_HAVE_32K_MODEL_ACCESS', 'true').lower() == 'true'

# goes under Modal 'slack' secret name
SLACK_CLIENT_ID = os.environ.get('SLACK_CLIENT_ID')
SLACK_CLIENT_SECRET = os.environ.get('SLACK_CLIENT_SECRET')
SLACK_APP_PAGE_URL = os.environ.get('SLACK_APP_PAGE_URL')
SLACK_APP_INSTALL_URL = os.environ.get('SLACK_APP_INSTALL_URL')

# goes under Modal 'anthropic' secret name
ANTHROPIC_API_KEY = os.environ.get('ANTHROPIC_API_KEY')

# goes under Modal 'mongodb' secret name
MONGODB_URI = os.environ.get('MONGODB_URI')

# goes under Modal 'redis_url' secret name (optional, can leave env var blank)
REDIS_URL = os.environ.get('REDIS_URL')
# deprecated: old logic transfer so upstream can use this
if not REDIS_URL:
    REDIS_URL = os.environ.get('redis_url')

# goes under Modal 'posthog' secret name (optional, can leave env var blank)
POSTHOG_API_KEY = os.environ.get('POSTHOG_API_KEY')

# goes under Modal 'highlight' secret name (optional, can leave env var blank)
HIGHLIGHT_API_KEY = os.environ.get('HIGHLIGHT_API_KEY')

SECONDARY_MODEL = "gpt-3.5-turbo-16k-0613"



import difflib


def generate_diff(old_code, new_code):
    diff = difflib.unified_diff(
        old_code.splitlines(keepends=True),
        new_code.splitlines(keepends=True)
    )
    return ''.join(diff)


def revert_whitespace_changes(original_file_str, modified_file_str):
    original_lines = original_file_str.splitlines()
    modified_lines = modified_file_str.splitlines()

    diff = difflib.SequenceMatcher(None, original_lines, modified_lines)

    final_lines = []
    for opcode in diff.get_opcodes():
        if opcode[0] == "equal" or opcode[0] == "replace":
            # If the lines are equal or replace (means the change is not whitespace only)
            # use original lines.
            final_lines.extend(original_lines[opcode[1]:opcode[2]])
        elif opcode[0] == "insert":
            # If the lines are inserted in the modified file, check if it's just whitespace changes
            # If it's just whitespace changes, ignore them.
            for line in modified_lines[opcode[3]:opcode[4]]:
                if line.strip() != "":
                    final_lines.append(line)

    return '\n'.join(final_lines)


def format_contents(file_contents, is_markdown=False):
    """
    Add arbitrary postprocessing here, this affects files and diffs
    """
    lines = file_contents.split('\n')

    if is_markdown:
        return '\n'.join(lines)

    # Handle small files
    if len(lines) <= 5:
        start_idx = 0
        end_idx = len(lines)
        for idx, line in enumerate(lines):
            if start_idx == 0 and line.strip().startswith('```'):
                start_idx = idx + 1
            if start_idx != 0 and line.strip().endswith('```'):
                end_idx = idx
        lines = lines[start_idx:end_idx]
        return '\n'.join(lines)

    first_three_lines = lines[:3]
    last_three_lines = lines[-3:]
    first_line_idx = 0
    last_line_idx = 3
    for idx, line in enumerate(first_three_lines):
        line = line.strip()
        if line.startswith('```'):
            first_line_idx = idx + 1
    for idx, line in enumerate(last_three_lines):
        line = line.strip()
        if line.endswith('```'):
            last_line_idx = idx
    first_three_lines = first_three_lines[first_line_idx:]
    last_three_lines = last_three_lines[:last_line_idx]

    lines = first_three_lines + lines[3:-3] + last_three_lines
    return '\n'.join(lines)


def generate_new_file(modify_file_response: str, old_file_content: str) -> str:
    import re

    result_file = ""
    old_file_lines = old_file_content.splitlines()

    # Extract content between <new_file> tags
    new_file = re.search(r".*?<new_file>\n(.*)\n<\/new_file>", modify_file_response, re.DOTALL).group(1)
    if "<copy_lines" not in new_file:
        return new_file

    # v5
    result = []
    lines = new_file.split('\n')
    for line_number, line in enumerate(lines):
        # Todo: make it support 1 number only
        matches = re.finditer(r"<copy_lines\s(\d+-\d+)/?>", line)
        copied_lines = False
        for match in matches:
            copied_lines = True
            start, end = match.group(1).split('-')
            start, end = int(start) - 1, int(end) - 1

            start = max(0, start)
            end = min(len(old_file_lines) - 1, end)

            replacements = old_file_lines[start:end + 1]
            replacements_str = '\n'.join(replacements)
            line = line.replace(match.group(0), replacements_str)

        # check if line was incorrectly duplicated
        append = True
        if not copied_lines:  # if bot generated, and line before is not bot generated
            if len(result) > 0:
                # Get last line in results
                last_group = result[-1]
                # last_line = last_group
                if '\n' in last_group:
                    last_line = last_group[last_group.rindex('\n') + 1:]  # if its multiple lines
                    # if last line is same is current line
                    if last_line == line:
                        append = False

        if append:
            result.append(line)
    result = '\n'.join(result)

    # Todo: v4 is inefficient; deprecated
    """
    first_section_idx = new_file.index("<copied>")
    if first_section_idx > 0:
        result_file += new_file[:first_section_idx]
        new_file = new_file[first_section_idx:] # remove the first section from new_file
    last_section_idx = new_file.rindex("</copied>")
    last_section = ""
    if last_section_idx < len(new_file) - 1:
        last_section = new_file[last_section_idx + len("</copied>"):]
        new_file = new_file[:last_section_idx + len("</copied>")] # remove the last section from new_file
    
    # Parse copied sections, first copying the content and then adding whatever is after the copied section
    for copied_section in copied_sections:
        if "-" in copied_section:
            start_line, end_line = copied_section.split("-")
        else: # <copied>num</copied>
            start_line = copied_sections
            end_line = start_line

        start_line = int(start_line) - 1 if int(start_line) - 1 > 0 else 0
        end_line = int(end_line) - 1
        # Check for duplicate lines
        k = 30
        result_file = join_contents_k(result_file, "\n".join(old_file_lines[start_line:end_line]), k)
        # TODO: Use replace first instead of .replace, since duplicated <copied> sections might cause faulty copy
        new_file = new_file.replace(f"<copied>{copied_section}</copied>\n", "")
        next_section_idx = new_file.index("<copied>") if "<copied>" in new_file else len(new_file)
        # Check for duplicate lines
        result_file = join_contents_k(result_file, new_file[:next_section_idx], k)
        new_file = new_file[next_section_idx:] # remove the first section from new_file
    
    return result_file + last_section
    """

    return result


def join_contents_k(first, second, k):
    """
    Join contents together removing k duplicate lines
    """
    first_lines = first.splitlines()
    second_lines = second.splitlines()
    for i in range(k, 0, -1):
        if len(first_lines) < k or len(second_lines) < k:
            continue
        if first_lines[-i:] == second_lines[:i]:
            return "\n".join(first_lines) + "\n" + "\n".join(second_lines[i:])
    return "\n".join(first_lines) + "\n" + "\n".join(second_lines)


def is_markdown(filename):
    return filename.endswith(".md") or filename.endswith(".rst") or filename.endswith(".txt")



import highlight_io
from loguru import logger
from posthog import Posthog

from sweepai.utils.config.server import POSTHOG_API_KEY, HIGHLIGHT_API_KEY

if POSTHOG_API_KEY is None:
    posthog = Posthog(project_api_key="none", disabled=True, host='https://app.posthog.com')
    logger.warning("Initialized an empty Posthog instance as POSTHOG_API_KEY is not present.")
else:
    posthog = Posthog(project_api_key=POSTHOG_API_KEY, host='https://app.posthog.com')

if HIGHLIGHT_API_KEY is None:
    H = None
else:
    H = highlight_io.H(HIGHLIGHT_API_KEY)
    logger.add(
        H.logging_handler,
        format="{message}",
        level="INFO",
        backtrace=True,
    )



'''
On Github ticket, get ChatGPT to deal with it
'''

# TODO: Add file validation

import traceback

import modal
import openai
from loguru import logger
from tabulate import tabulate

from sweepai.core.entities import Snippet, NoFilesException
from sweepai.core.sweep_bot import SweepBot, MaxTokensExceeded
from sweepai.core.prompts import issue_comment_prompt
from sweepai.handlers.create_pr import create_pr, create_config_pr, safe_delete_sweep_branch
from sweepai.handlers.on_comment import on_comment
from sweepai.handlers.on_review import review_pr
from sweepai.utils.chat_logger import ChatLogger, discord_log_error
from sweepai.utils.config.client import SweepConfig
from sweepai.utils.config.server import PREFIX, DB_MODAL_INST_NAME, UTILS_MODAL_INST_NAME, OPENAI_API_KEY, \
    GITHUB_BOT_TOKEN, \
    GITHUB_BOT_USERNAME
from sweepai.utils.event_logger import posthog
from sweepai.utils.github_utils import get_github_client, search_snippets
from sweepai.utils.prompt_constructor import HumanMessagePrompt

github_access_token = GITHUB_BOT_TOKEN
openai.api_key = OPENAI_API_KEY

update_index = modal.Function.lookup(DB_MODAL_INST_NAME, "update_index")

sep = "\n---\n"
bot_suffix_starring = "⭐ If you are enjoying Sweep, please [star our repo](https://github.com/sweepai/sweep) so more people can hear about us!"
bot_suffix = f"\n{sep}I'm a bot that handles simple bugs and feature requests but I might make mistakes. Please be kind!"
discord_suffix = f'\n<sup>[Join Our Discord](https://discord.com/invite/sweep-ai)'

stars_suffix = "⭐ In the meantime, consider [starring our repo](https://github.com/sweepai/sweep) so more people can hear about us!"

collapsible_template = '''
<details>
  <summary>{summary}</summary>

  {body}
</details>
'''

chunker = modal.Function.lookup(UTILS_MODAL_INST_NAME, "Chunking.chunk")

num_of_snippets_to_query = 30
total_number_of_snippet_tokens = 15_000
num_full_files = 2
num_extended_snippets = 2


def post_process_snippets(snippets: list[Snippet], max_num_of_snippets: int = 5):
    for snippet in snippets[:num_full_files]:
        snippet = snippet.expand()

    # snippet fusing
    i = 0
    while i < len(snippets):
        j = i + 1
        while j < len(snippets):
            if snippets[i] ^ snippets[j]:  # this checks for overlap
                snippets[i] = snippets[i] | snippets[j]  # merging
                snippets.pop(j)
            else:
                j += 1
        i += 1

    # truncating snippets based on character length
    result_snippets = []
    total_length = 0
    for snippet in snippets:
        total_length += len(snippet.get_snippet())
        if total_length > total_number_of_snippet_tokens * 5:
            break
        result_snippets.append(snippet)
    return result_snippets[:max_num_of_snippets]


def on_ticket(
        title: str,
        summary: str,
        issue_number: int,
        issue_url: str,
        username: str,
        repo_full_name: str,
        repo_description: str,
        installation_id: int,
        comment_id: int = None
):
    # Check if the title starts with "sweep" or "sweep: " and remove it
    if title.lower().startswith("sweep: "):
        title = title[7:]
    elif title.lower().startswith("sweep "):
        title = title[6:]

    # Flow:
    # 1. Get relevant files
    # 2: Get human message
    # 3. Get files to change
    # 4. Get file changes
    # 5. Create PR

    organization, repo_name = repo_full_name.split("/")
    metadata = {
        "issue_url": issue_url,
        "repo_name": repo_name,
        "repo_description": repo_description,
        "username": username,
        "installation_id": installation_id,
        "function": "on_ticket",
        "mode": PREFIX,
    }
    posthog.capture(username, "started", properties=metadata)

    g = get_github_client(installation_id)

    logger.info(f"Getting repo {repo_full_name}")
    repo = g.get_repo(repo_full_name)
    current_issue = repo.get_issue(number=issue_number)
    if current_issue.state == 'closed':
        posthog.capture(username, "issue_closed", properties=metadata)
        return {"success": False, "reason": "Issue is closed"}
    item_to_react_to = current_issue.get_comment(comment_id) if comment_id else current_issue
    replies_text = ""
    comments = list(current_issue.get_comments())
    if comment_id:
        logger.info(f"Replying to comment {comment_id}...")
        replies_text = "\nComments:\n" + "\n".join(
            [
                issue_comment_prompt.format(
                    username=comment.user.login,
                    reply=comment.body,
                ) for comment in comments if comment.user.type == "User"
            ]
        )

    chat_logger = ChatLogger({
        'repo_name': repo_name,
        'title': title,
        'summary': summary + replies_text,
        "issue_number": issue_number,
        "issue_url": issue_url,
        "username": username,
        "repo_full_name": repo_full_name,
        "repo_description": repo_description,
        "installation_id": installation_id,
        "comment_id": comment_id,
    })

    # Check if branch was already created for this issue
    preexisting_branch = None
    prs = repo.get_pulls(state='open', sort='created', base=SweepConfig.get_branch(repo))
    for pr in prs:
        # Check if this issue is mentioned in the PR, and pr is owned by bot
        # This is done in create_pr, (pr_description = ...)
        if pr.user.login == GITHUB_BOT_USERNAME and f'Fixes #{issue_number}.\n' in pr.body:
            success = safe_delete_sweep_branch(pr, repo)

    # Add emojis
    eyes_reaction = item_to_react_to.create_reaction("eyes")
    # If SWEEP_BOT reacted to item_to_react_to with "rocket", then remove it.
    reactions = item_to_react_to.get_reactions()
    for reaction in reactions:
        if reaction.content == "rocket" and reaction.user.login == GITHUB_BOT_USERNAME:
            item_to_react_to.delete_reaction(reaction.id)

    # Creates progress bar ASCII for 0-5 states
    progress_headers = [
        None,
        "Step 1: 🔍 Code Search",
        "Step 2: 🧐 Snippet Analysis",
        "Step 3: 📝 Planning",
        "Step 4: ⌨️ Coding",
        "Step 5: 🔁 Code Review"
    ]

    config_pr_url = None

    def get_comment_header(index, errored=False, pr_message=""):
        config_pr_message = (
            "\n" + f"* Install Sweep Configs: [Pull Request]({config_pr_url})" if config_pr_url is not None else "")
        if index < 0: index = 0
        if index == 5:
            return pr_message + config_pr_message
        index *= 20
        index = min(100, index)
        if errored:
            return f"![{index}%](https://progress-bar.dev/{index}/?&title=Errored&width=600)"
        return f"![{index}%](https://progress-bar.dev/{index}/?&title=Progress&width=600)" + (
            "\n" + stars_suffix + config_pr_message if index != -1 else "")

    # Find the first comment made by the bot
    issue_comment = None
    is_paying_user = chat_logger.is_paying_user()
    tickets_allocated = 60 if is_paying_user else 3
    ticket_count = max(tickets_allocated - chat_logger.get_ticket_count(), 0)
    use_faster_model = chat_logger.use_faster_model()
    payment_message = f"To create this ticket, I used {'gpt-3.5. ' if use_faster_model else 'gpt-4. '}You have {ticket_count} gpt-4 tickets left." + (" For more gpt-4 tickets, visit [our payment portal.](https://buy.stripe.com/fZe03512h99u0AE6os)" if not is_paying_user else "")
    first_comment = f"{get_comment_header(0)}\n{sep}I am currently looking into this ticket!. I will update the progress of the ticket in this comment. I am currently searching through your code, looking for relevant snippets.\n{sep}## {progress_headers[1]}\nWorking on it...{bot_suffix}{discord_suffix}"
    for comment in comments:
        if comment.user.login == GITHUB_BOT_USERNAME:
            issue_comment = comment
            issue_comment.edit(first_comment)
            break
    if issue_comment is None:
        issue_comment = current_issue.create_comment(first_comment)

    # Comment edit function
    past_messages = {}
    current_index = {}

    def edit_sweep_comment(message: str, index: int, pr_message=""):
        nonlocal current_index
        # -1 = error, -2 = retry
        # Only update the progress bar if the issue generation errors.
        errored = (index == -1)
        if index >= 0:
            past_messages[index] = message
            current_index = index

        agg_message = None
        # Include progress history
        # index = -2 is reserved for
        for i in range(current_index + 2):  # go to next header (for Working on it... text)
            if i == 0 or i >= len(progress_headers): continue  # skip None header
            header = progress_headers[i]
            if header is not None:
                header = "## " + header + "\n"
            else:
                header = "No header\n"
            msg = header + (past_messages.get(i) or "Working on it...")
            if agg_message is None:
                agg_message = msg
            else:
                agg_message = agg_message + f"\n{sep}" + msg

        suffix = bot_suffix + discord_suffix
        if errored:
            agg_message = "## ❌ Unable to Complete PR" + '\n' + message + "\nIf you would like to report this bug, please join our **[Discord](https://discord.com/invite/sweep-ai)**."
            suffix = bot_suffix # don't include discord suffix for error messages

        # Update the issue comment
        issue_comment.edit(f"{get_comment_header(current_index, errored, pr_message)}\n{sep}{agg_message}{suffix}")

    def log_error(error_type, exception):
        content = f"**{error_type} Error**\n{username}: {issue_url}\n```{exception}```"
        discord_log_error(content)

    def fetch_file_contents_with_retry():
        retries = 3
        error = None
        for i in range(retries):
            try:
                logger.info(f"Fetching relevant files for the {i}th time...")
                return search_snippets(
                    repo,
                    f"{title}\n{summary}\n{replies_text}",
                    num_files=num_of_snippets_to_query,
                    branch=None,
                    installation_id=installation_id,
                )
            except Exception as e:
                error = e
                continue
        posthog.capture(
            username, "fetching_failed", properties={"error": error, **metadata}
        )
        raise error

    logger.info("Fetching relevant files...")
    try:
        snippets, tree = fetch_file_contents_with_retry()
        assert len(snippets) > 0
    except Exception as e:
        trace = traceback.format_exc()
        logger.error(e)
        logger.error(trace)
        edit_sweep_comment(
            "It looks like an issue has occured around fetching the files. Perhaps the repo has not been initialized: try removing this repo and adding it back. I'll try again in a minute. If this error persists contact team@sweep.dev.",
            -1
        )
        log_error("File Fetch", str(e) + "\n" + traceback.format_exc())
        raise e

    snippets = post_process_snippets(snippets)

    snippets = post_process_snippets(snippets,
                                     max_num_of_snippets=2 if use_faster_model else 5)

    human_message = HumanMessagePrompt(
        repo_name=repo_name,
        issue_url=issue_url,
        username=username,
        repo_description=repo_description,
        title=title,
        summary=summary + replies_text,
        snippets=snippets,
        tree=tree,  # TODO: Anything in repo tree that has something going through is expanded
    )

    sweep_bot = SweepBot.from_system_message_content(
        human_message=human_message, repo=repo, is_reply=bool(comments), chat_logger=chat_logger
    )

    # Check repository for sweep.yml file.
    sweep_yml_exists = False
    for content_file in repo.get_contents(""):
        if content_file.name == "sweep.yaml":
            sweep_yml_exists = True
            break

    # If sweep.yaml does not exist, then create a new PR that simply creates the sweep.yaml file.
    if not sweep_yml_exists:
        try:
            logger.info("Creating sweep.yaml file...")
            config_pr = create_config_pr(sweep_bot)
            config_pr_url = config_pr.html_url
            edit_sweep_comment(message="", index=-2)
        except Exception as e:
            logger.error("Failed to create new branch for sweep.yaml file.\n", e, traceback.format_exc())
    else:
        logger.info("sweep.yaml file already exists.")

    sweepbot_retries = 3
    try:
        for i in range(sweepbot_retries):
            # ANALYZE SNIPPETS
            if sweep_bot.model == "gpt-4-32k-0613":
                logger.info("CoT retrieval...")
                sweep_bot.cot_retrieval()
            else:
                logger.info("Did not execute CoT retrieval...")

            newline = '\n'
            edit_sweep_comment(
                "I found the following snippets in your repository. I will now analyze these snippets and come up with a plan."
                + "\n\n"
                + collapsible_template.format(
                    summary="Some code snippets I looked at (click to expand). If some file is missing from here, you can mention the path in the ticket description.",
                    body="\n".join(
                        [
                            f"https://github.com/{organization}/{repo_name}/blob/{repo.get_commits()[0].sha}/{snippet.file_path}#L{max(snippet.start, 1)}-L{min(snippet.end, snippet.content.count(newline))}\n"
                            for snippet in snippets
                        ]
                    ),
                ),
                1
            )

            # COMMENT ON ISSUE
            # TODO: removed issue commenting here
            logger.info("Fetching files to modify/create...")
            file_change_requests = sweep_bot.get_files_to_change()
            file_change_requests = sweep_bot.validate_file_change_requests(file_change_requests)
            table = tabulate(
                [[f"`{file_change_request.filename}`", file_change_request.instructions] for file_change_request in
                 file_change_requests],
                headers=["File Path", "Proposed Changes"],
                tablefmt="pipe"
            )
            edit_sweep_comment(
                "From looking through the relevant snippets, I decided to make the following modifications:\n\n" + table + "\n\n",
                2
            )

            # CREATE PR METADATA
            logger.info("Generating PR...")
            pull_request = sweep_bot.generate_pull_request()
            pull_request_content = pull_request.content.strip().replace("\n", "\n>")
            pull_request_summary = f"**{pull_request.title}**\n`{pull_request.branch_name}`\n>{pull_request_content}\n"
            edit_sweep_comment(
                f"I have created a plan for writing the pull request. I am now working my plan and coding the required changes to address this issue. Here is the planned pull request:\n\n{pull_request_summary}",
                3
            )

            # WRITE PULL REQUEST
            logger.info("Making PR...")
            response = create_pr(file_change_requests, pull_request, sweep_bot, username, installation_id, issue_number)
            if not response or not response["success"]: raise Exception("Failed to create PR")
            pr = response["pull_request"]
            current_issue.create_reaction("rocket")
            edit_sweep_comment(
                "I have finished coding the issue. I am now reviewing it for completeness.",
                4
            )

            try:
                current_issue.delete_reaction(eyes_reaction.id)
            except:
                pass
            try:
                # CODE REVIEW
                changes_required, review_comment = review_pr(repo=repo, pr=pr, issue_url=issue_url, username=username,
                                                             repo_description=repo_description, title=title,
                                                             summary=summary, replies_text=replies_text, tree=tree)
                logger.info(f"Addressing review comment {review_comment}")
                if changes_required:
                    on_comment(repo_full_name=repo_full_name,
                               repo_description=repo_description,
                               comment=review_comment,
                               username=username,
                               installation_id=installation_id,
                               pr_path=None,
                               pr_line_position=None,
                               pr_number=pr.number)
            except Exception as e:
                logger.error(traceback.format_exc())
                logger.error(e)

            # Completed code review
            edit_sweep_comment(
                "Success! 🚀",
                5,
                pr_message=f"## Here's the PR! [https://github.com/{repo_full_name}/pull/{pr.number}](https://github.com/{repo_full_name}/pull/{pr.number}).\n{payment_message}",
            )

            break
    except MaxTokensExceeded as e:
        logger.info("Max tokens exceeded")
        log_error("Max Tokens Exceeded", str(e) + "\n" + traceback.format_exc())
        if chat_logger.is_paying_user():
            edit_sweep_comment(f"Sorry, I could not edit `{e.filename}` as this file is too long. We are currently working on improved file streaming to address this issue.\n", -1)
        else:
            edit_sweep_comment(f"Sorry, I could not edit `{e.filename}` as this file is too long.\n\nIf this file is incorrect, please describe the desired file in the prompt. However, if you would like to edit longer files, consider upgrading to [Sweep Pro](https://sweep.dev/) for longer context lengths.\n", -1)
        raise e
    except NoFilesException as e:
        logger.info("No files to change.")
        log_error("No Files to Change", str(e) + "\n" + traceback.format_exc())
        edit_sweep_comment("Sorry, I could find any appropriate files to edit to address this issue. If this is a mistake, please provide more context and I will retry!", -1)
        raise e
    except openai.error.InvalidRequestError as e:
        logger.error(traceback.format_exc())
        logger.error(e)
        edit_sweep_comment(
            "I'm sorry, but it looks our model has ran out of context length. We're trying to make this happen less, but one way to mitigate this is to code smaller files. If this error persists contact team@sweep.dev.",
            -1
        )
        log_error("Context Length", str(e) + "\n" + traceback.format_exc())
        posthog.capture(
            username,
            "failed",
            properties={
                "error": str(e),
                "reason": "Invalid request error / context length",
                **metadata,
            },
        )
        raise e
    except Exception as e:
        logger.error(traceback.format_exc())
        logger.error(e)
        edit_sweep_comment(
            "I'm sorry, but it looks like an error has occured. Try removing and re-adding the sweep label. If this error persists contact team@sweep.dev.",
            -1
        )
        log_error("Workflow", str(e) + "\n" + traceback.format_exc())
        posthog.capture(
            username,
            "failed",
            properties={"error": str(e), "reason": "Generic error", **metadata},
        )
        raise e
    else:
        try:
            item_to_react_to.delete_reaction(eyes_reaction.id)
        except:
            pass
        item_to_react_to.create_reaction("rocket")

    posthog.capture(username, "success", properties={**metadata})
    logger.info("on_ticket success")
    return {"success": True}



from loguru import logger
from pydantic import BaseModel

from sweepai.core.prompts import (
    human_message_prompt,
    human_message_prompt_comment,
    human_message_review_prompt,
    diff_section_prompt,
    review_follow_up_prompt,
    final_review_prompt,
    comment_line_prompt
)


class HumanMessagePrompt(BaseModel):
    repo_name: str
    issue_url: str
    username: str
    title: str
    summary: str
    snippets: list
    tree: str
    repo_description: str = ""

    def delete_file(self, file_path):
        # super unoptimized
        if file_path in [snippet.file_path for snippet in self.snippets]:
            for snippet in self.snippets:
                if snippet.file_path == file_path:
                    self.snippets.remove(snippet)

    def get_relevant_directories(self):
        deduped_paths = []
        for snippet in self.snippets:
            if snippet.file_path not in deduped_paths:
                deduped_paths.append(snippet.file_path)
        return "\n".join(deduped_paths)

    def render_snippets(self):
        return "\n".join([snippet.xml for snippet in self.snippets])

    def construct_prompt(self):
        human_messages = [{'role': msg['role'], 'content': msg['content'].format(
            repo_name=self.repo_name,
            issue_url=self.issue_url,
            username=self.username,
            repo_description=self.repo_description,
            tree=self.tree,
            title=self.title,
            description=self.summary if self.summary else "No description provided.",
            relevant_snippets=self.render_snippets(),
            relevant_directories=self.get_relevant_directories(),
        ), 'key': msg.get('key')} for msg in human_message_prompt]
        return human_messages


class HumanMessagePromptReview(HumanMessagePrompt):
    pr_title: str
    pr_message: str = ""
    diffs: list

    def format_diffs(self):
        formatted_diffs = []
        for file_name, new_file_contents, old_file_contents, file_patch in self.diffs:
            format_diff = diff_section_prompt.format(
                diff_file_path=file_name,
                new_file_content=new_file_contents.rstrip("\n"),
                previous_file_content=old_file_contents.rstrip("\n"),
                diffs=file_patch
            )
            formatted_diffs.append(format_diff)
        return "\n".join(formatted_diffs)

    def construct_prompt(self):
        human_messages = [{'role': msg['role'], 'content': msg['content'].format(
            repo_name=self.repo_name,
            issue_url=self.issue_url,
            username=self.username,
            repo_description=self.repo_description,
            tree=self.tree,
            title=self.title,
            description=self.summary,
            relevant_snippets=self.render_snippets(),
            relevant_directories=self.get_relevant_directories(),
            diffs=self.format_diffs(),
            pr_title=self.pr_title,
            pr_message=self.pr_message,
        )} for msg in human_message_review_prompt]

        return human_messages


class HumanMessageReviewFollowup(BaseModel):
    diff: tuple

    def construct_prompt(self):
        file_name, new_file_contents, old_file_contents, file_patch = self.diff
        format_diff = diff_section_prompt.format(
            diff_file_path=file_name,
            new_file_content=new_file_contents.rstrip("\n"),
            previous_file_content=old_file_contents.rstrip("\n"),
            diffs=file_patch
        )
        return review_follow_up_prompt + format_diff


class HumanMessageCommentPrompt(HumanMessagePrompt):
    comment: str
    diffs: list
    pr_file_path: str | None
    pr_line: str | None

    def format_diffs(self):
        formatted_diffs = []
        for file_name, new_file_contents, old_file_contents, file_patch in self.diffs:
            format_diff = diff_section_prompt.format(
                diff_file_path=file_name,
                new_file_content=new_file_contents.rstrip("\n"),
                previous_file_content=old_file_contents.rstrip("\n"),
                diffs=file_patch
            )
            formatted_diffs.append(format_diff)
        return "\n".join(formatted_diffs)

    def construct_prompt(self):
        human_messages = [{'role': msg['role'], 'content': msg['content'].format(
            comment=self.comment,
            repo_name=self.repo_name,
            repo_description=self.repo_description if self.repo_description else "",
            diff=self.format_diffs(),
            issue_url=self.issue_url,
            username=self.username,
            title=self.title,
            tree=self.tree,
            description=self.summary if self.summary else "No description provided.",
            relevant_directories=self.get_relevant_directories(),
            relevant_snippets=self.render_snippets()
        )} for msg in human_message_prompt_comment]

        if self.pr_file_path and self.pr_line:
            logger.info(f"Review Comment {self.comment}")
            human_messages.append({'role': 'user', 'content': comment_line_prompt.format(
                pr_file_path=self.pr_file_path,
                pr_line=self.pr_line
            )})
        else:
            logger.info(f"General Comment {self.comment}")

        return human_messages


class HumanMessageFinalPRComment(BaseModel):
    summarization_replies: list

    def construct_prompt(self):
        final_review = final_review_prompt.format(
            file_summaries="\n".join(self.summarization_replies)
        )
        return final_review



"""
Proxy for the UI.
"""

from datetime import datetime
import json
from typing import Any

import fastapi
import modal
from fastapi import FastAPI
from fastapi.responses import StreamingResponse
from github import Github
from loguru import logger
from pydantic import BaseModel
from pymongo import MongoClient

from sweepai.app.config import SweepChatConfig
from sweepai.core.chat import ChatGPT
from sweepai.core.entities import FileChangeRequest, Function, Message, PullRequest, Snippet
from sweepai.core.prompts import gradio_system_message_prompt, gradio_user_prompt
from sweepai.core.sweep_bot import SweepBot
from sweepai.utils.config.client import SweepConfig
from sweepai.utils.config.server import MONGODB_URI, PREFIX, DB_MODAL_INST_NAME, API_MODAL_INST_NAME, BOT_TOKEN_NAME
from sweepai.utils.event_logger import posthog
from sweepai.utils.github_utils import get_github_client, get_installation_id

get_relevant_snippets = modal.Function.from_name(DB_MODAL_INST_NAME, "get_relevant_snippets")

stub = modal.Stub(PREFIX + "-ui")
image = (
    modal.Image.debian_slim()
    .apt_install("git")
    .pip_install(
        "loguru",
        "tqdm",
        "posthog",
        "openai",
        "anthropic",
        "highlight-io",
        "PyGithub",
        "GitPython",
        "config-path",
        "pyyaml",
        "pymongo",
    )
)
secrets = [
    modal.Secret.from_name(BOT_TOKEN_NAME),
    modal.Secret.from_name("github"),
    modal.Secret.from_name("openai-secret"),
    modal.Secret.from_name("posthog"),
    modal.Secret.from_name("highlight"),
    modal.Secret.from_name("mongodb"),
]

FUNCTION_SETTINGS = {
    "image": image,
    "secrets": secrets,
    "timeout": 15 * 60,
    "keep_warm": 1
}


@stub.function(**FUNCTION_SETTINGS)
@modal.asgi_app(label=PREFIX + "-ui")
def _asgi_app():
    app = FastAPI()

    def verify_user(request: SweepChatConfig) -> bool:
        try:
            github_user_client = Github(request.github_pat)
            assert github_user_client.get_user().login == request.github_username
        except Exception as e:
            logger.warning(e)
            raise fastapi.HTTPException(status_code=403, detail="You do not have access to this repo")
        return True

    def verify_config(request: SweepChatConfig) -> bool:
        try:
            github_user_client = Github(request.github_pat)
            repo = github_user_client.get_repo(request.repo_full_name)
            assert repo
        except Exception as e:
            logger.warning(e)
            raise fastapi.HTTPException(status_code=403, detail="You do not have access to this repo")
        return True

    @app.post("/user_info")
    def user_info(request: SweepChatConfig) -> dict:
        assert verify_user(request)

        metadata = {
            "function": "ui_user_info",
            "repo_full_name": request.repo_full_name,
            "organization": (request.repo_full_name or "/").split("/")[0],
            "username": request.github_username,
            "installation_id": request.installation_id,
            "mode": PREFIX,
        }

        posthog.capture(request.github_username, "started", properties=metadata)

        current_month = datetime.utcnow().strftime('%m/%Y') 

        client = MongoClient(MONGODB_URI, serverSelectionTimeoutMS=5000, socketTimeoutMS=5000)
        db = client['llm']
        ticket_collection = db['tickets']
        user = ticket_collection.find_one({'username': request.github_username})
        is_paying_user = user.get('is_paying_user', False) if user else False

        result = ticket_collection.aggregate([
            {'$match': {'username': request.github_username}},
            {'$project': {current_month: 1, '_id': 0}}
        ])
        result_list = list(result)
        ticket_count = result_list[0].get(current_month, 0) if len(result_list) > 0 else 0
        logger.info(f'Ticket Count for {request.github_username} {ticket_count}')

        posthog.capture(request.github_username, "success", properties=metadata)

        return {
            "is_paying_user": is_paying_user,
            "remaining_tickets": max((60 if is_paying_user else 5) - ticket_count, 0),
        }

    @app.post("/installation_id")
    def installation_id(request: SweepChatConfig) -> dict:
        # first check if user has access to the repo
        assert verify_config(request)

        metadata = {
            "function": "ui_installation_id",
            "repo_full_name": request.repo_full_name,
            "organization": request.repo_full_name.split("/")[0],
            "username": request.github_username,
            "installation_id": request.installation_id,
            "mode": PREFIX,
        }

        posthog.capture(request.github_username, "started", properties=metadata)

        try:
            organization, _repo_name = request.repo_full_name.split("/")
            installation_id = get_installation_id(organization)
            assert installation_id
        except Exception as e:
            logger.warning(e)
            posthog.capture(request.github_username, "failed", properties={"error": str(e), **metadata})
            raise fastapi.HTTPException(status_code=403,
                                        detail="Sweep app is not installed on this repo. To install it, go to https://github.com/apps/sweep-ai")

        posthog.capture(request.github_username, "success", properties=metadata)

        return {"installation_id": installation_id}

    class SearchRequest(BaseModel):
        query: str
        config: SweepChatConfig
        n_results: int = 5

    @app.post("/search")
    def search(request: SearchRequest) -> list[Snippet]:
        logger.info("Searching for snippets...")
        get_relevant_snippets = modal.Function.lookup(DB_MODAL_INST_NAME, "get_relevant_snippets")

        assert verify_config(request.config)

        metadata = {
            "function": "ui_search",
            "repo_full_name": request.config.repo_full_name,
            "organization": request.config.repo_full_name.split("/")[0],
            "username": request.config.github_username,
            "installation_id": request.config.installation_id,
            "mode": PREFIX,
        }

        posthog.capture(request.config.github_username, "started", properties=metadata)

        try:
            snippets: list[Snippet] = get_relevant_snippets.call(
                request.config.repo_full_name,
                request.query,
                n_results=request.n_results,
                installation_id=request.config.installation_id
            )
            g = get_github_client(request.config.installation_id)
            repo = g.get_repo(request.config.repo_full_name)
            for snippet in snippets:
                try:
                    snippet.content = repo.get_contents(snippet.file_path,
                                                        SweepConfig.get_branch(repo)).decoded_content.decode("utf-8")
                except Exception:
                    logger.error(snippet)
        except Exception as e:
            posthog.capture(request.config.github_username, "failed", properties={"error": str(e), **metadata})
            raise e

        posthog.capture(request.config.github_username, "success", properties=metadata)
        return snippets

    class CreatePRRequest(BaseModel):
        # proposed PR information
        file_change_requests: list[tuple[str, str]]
        pull_request: PullRequest

        # state information
        messages: list[tuple[str | None, str | None]]
        snippets: list[Snippet] = []

        config: SweepChatConfig

    @app.post("/create_pr")
    def create_pr(request: CreatePRRequest):
        assert verify_config(request.config)

        g = get_github_client(request.config.installation_id)
        repo = g.get_repo(request.config.repo_full_name)

        metadata = {
            "function": "ui_create_pr",
            "repo_full_name": request.config.repo_full_name,
            "organization": request.config.repo_full_name.split("/")[0],
            "username": request.config.github_username,
            "installation_id": request.config.installation_id,
            "mode": PREFIX,
        }

        posthog.capture(request.config.github_username, "started", properties=metadata)

        try:
            create_pr_func = modal.Function.lookup(API_MODAL_INST_NAME, "create_pr")
            system_message = gradio_system_message_prompt.format(
                snippets="\n".join(
                    [snippet.denotation + f"\n```{snippet.get_snippet()}```" for snippet in request.snippets]),
                repo_name=request.config.repo_full_name,
                repo_description=repo.description
            )

            def file_exists(file_path: str) -> bool:
                try:
                    repo.get_contents(file_path, SweepConfig.get_branch(repo))
                    return True
                except Exception:
                    return False

            results = create_pr_func.call(
                [FileChangeRequest(
                    filename=item[0],
                    instructions=item[1],
                    change_type="modify" if file_exists(item[0]) else "create",  # TODO update this
                ) for item in request.file_change_requests],
                request.pull_request,
                SweepBot(
                    repo=repo,
                    messages=[Message(role="system", content=system_message, key="system")] +
                             [Message.from_tuple(message) for message in request.messages],
                ),
                request.config.github_username,
                installation_id=request.config.installation_id,
                issue_number=None,
            )
            generated_pull_request = results["pull_request"]
            print(generated_pull_request)
        except Exception as e:
            posthog.capture(request.config.github_username, "failed", properties={
                "error": str(e),
                **metadata
            })
            raise e

        posthog.capture(request.config.github_username, "success", properties=metadata)
        return {
            "html_url": generated_pull_request.html_url,
        }

    class ChatRequest(BaseModel):
        messages: list[tuple[str | None, str | None]]
        snippets: list[Snippet]
        config: SweepChatConfig
        do_add_plan: bool = False
        functions: list[Function] = []
        function_call: Any = "auto"

    @app.post("/chat")
    def chat(
            request: ChatRequest,
    ) -> str:
        assert verify_config(request.config)

        messages = [Message.from_tuple(message) for message in request.messages]
        chatgpt = ChatGPT(messages=messages[:-1])
        result = chatgpt.chat(messages[-1].content, model="gpt-4-0613")
        return result

    @app.post("/chat_stream")
    def chat_stream(request: ChatRequest):
        assert verify_config(request.config)
        metadata = {
            "function": "ui_chat_stream",
            "repo_full_name": request.config.repo_full_name,
            "organization": request.config.repo_full_name.split("/")[0],
            "username": request.config.github_username,
            "installation_id": request.config.installation_id,
            "mode": PREFIX,
        }

        posthog.capture(request.config.github_username, "started", properties=metadata)
        try:
            messages = [Message.from_tuple(message) for message in request.messages]
            system_message = gradio_system_message_prompt.format(
                snippets="\n".join(
                    [snippet.denotation + f"\n```{snippet.get_snippet()}```" for snippet in request.snippets]),
                repo_name=request.config.repo_full_name,
                repo_description=""  # TODO: fill this
            )
            chatgpt = ChatGPT(messages=[Message(role="system", content=system_message, key="system")] + messages[:-1])
            if request.do_add_plan:
                chatgpt.messages[-1].content += gradio_user_prompt
        except Exception as e:
            posthog.capture(request.config.github_username, "failed", properties={"error": str(e), **metadata})
            raise e

        def stream_chat():
            for chunk in chatgpt.chat_stream(messages[-1].content, model="gpt-4-0613", functions=request.functions,
                                             function_call=request.function_call):
                yield json.dumps(chunk)
            posthog.capture(request.config.github_username, "success", properties=metadata)

        return StreamingResponse(
            stream_chat(),
            media_type="text/event-stream"
        )

    return app



import json
import os
import shutil
import tempfile
import re
import webbrowser

import gradio as gr
from git import Repo
from github import Github
from loguru import logger

from sweepai.app.api_client import APIClient, create_pr_function, create_pr_function_call
from sweepai.app.config import State, SweepChatConfig
from sweepai.core.entities import Snippet
from sweepai.utils.config.client import SweepConfig

config = SweepChatConfig.load()

api_client = APIClient(config=config)

pr_summary_template = '''⏳ I'm creating the following PR...

**{title}**
{summary}

Here is my plan:
{plan}

Reply to propose changes to the plan.'''

print("Getting list of repos...")
github_client = Github(config.github_pat)
repos = list(github_client.get_user().get_repos())
print("Done.")

css = '''
footer {
    visibility: hidden;
}
pre, code {
    white-space: pre-wrap !important;
    word-break: break-all !important;
}
#snippets {
    height: 400px;
    overflow-y: scroll;
}
#message_box > label > span {
    display: none;
}
'''


def get_files_recursively(root_path, path=''):
    files = []
    path_to_contents = {}

    if path == '.git':
        return files, path_to_contents

    current_dir = os.path.join(root_path, path)
    entries = os.listdir(current_dir)

    for entry in entries:
        entry_path = os.path.join(current_dir, entry)

        if os.path.isfile(entry_path):
            try:
                with open(entry_path, 'r', encoding="utf-8", errors="ignore") as file:
                    contents = file.read()
                path_to_contents[entry_path[len(root_path) + 1:]] = contents
                files.append(entry_path[len(root_path) + 1:])
            except UnicodeDecodeError as e:
                logger.warning(f"Received warning {e}, skipping...")
                continue
        elif os.path.isdir(entry_path):
            subfiles, subpath_to_contents = get_files_recursively(root_path, os.path.join(path, entry))
            files.extend(subfiles)
            path_to_contents.update(subpath_to_contents)

    return files, path_to_contents


def get_installation_id(repo_full_name):
    config.repo_full_name = repo_full_name
    api_client.config = config
    installation_id = api_client.get_installation_id()
    return installation_id


path_to_contents = {}


def get_files(repo_full_name):
    global path_to_contents
    global repo
    if repo_full_name is None:
        all_files = []
    else:
        # Make sure repo is added to Sweep before checking all recursive files
        try:
            installation_id = get_installation_id(repo_full_name)
            assert installation_id
        except:
            return []
        repo = github_client.get_repo(repo_full_name)
        branch_name = SweepConfig.get_branch(repo)
        repo_url = f"https://x-access-token:{config.github_pat}@github.com/{repo_full_name}.git"
        try:
            repo_dir = os.path.join(tempfile.gettempdir(), repo_full_name)
            if os.path.exists(repo_dir):
                git_repo = Repo(repo_dir)
            else:
                git_repo = Repo.clone_from(repo_url, repo_dir)
            git_repo.git.checkout(branch_name)
            git_repo.remotes.origin.pull()
        except Exception as e:
            logger.warning(f"Git pull failed with error {e}, deleting cache and recloning...")
            shutil.rmtree(repo_dir)
            git_repo = Repo.clone_from(repo_url, repo_dir)
            git_repo.git.checkout(branch_name)
            git_repo.remotes.origin.pull()
        all_files, path_to_contents = get_files_recursively(repo_dir)
    return all_files


def get_files_update(*args):
    global repo
    if len(args) > 0:
        repo = args[0]
    else:
        repo = config.repo_full_name
    return gr.Dropdown.update(choices=get_files(repo))


def parse_response(raw_response: str) -> tuple[str, list[tuple[str, str]]]:
    if "Plan:" not in raw_response:
        response, raw_plan = raw_response, ""
    else:
        response, raw_plan = raw_response.split("Plan:", 1)
    if response.startswith("Response:"):
        response = response[len("Response:"):]
    plan = [(line[:line.find(":")].strip(), line[line.find(":") + 1:].strip()) for line in raw_plan.split("\n*") if
            line]
    return response, plan

try:
    user_info = api_client.get_user_info()
except Exception as e:
    logger.warning(e)
    user_info = {"is_paying_user": False, "remaining_tickets": 0}
global_state = config.state

with gr.Blocks(theme=gr.themes.Soft(), title="Sweep Chat", css=css) as demo:
    print("Launching gradio!")
    with gr.Row():
        with gr.Column(scale=2):
            repo_full_name = gr.Dropdown(choices=[repo.full_name for repo in repos], label="Repo full name",
                                         value=lambda: config.repo_full_name or "")
        print("Indexing files...")
        with gr.Column(scale=4):
            file_names = gr.Dropdown(choices=get_files(config.repo_full_name), multiselect=True, label="Files",
                                     value=lambda: global_state.file_paths)
        print("Indexed files!")
        repo_full_name.change(get_files_update, repo_full_name, file_names)
        with gr.Column(scale=1):
            restart_button = gr.Button("Restart")

    with gr.Row():
        with gr.Column(scale=2):
            chatbot = gr.Chatbot(height=400, value=lambda: global_state.chat_history)
        with gr.Column():
            with gr.Row():
                snippets_text = gr.Markdown(value=lambda: global_state.snippets_text, elem_id="snippets")

    with gr.Row():
        plan = gr.List(
            value=[[filename + ": " + instructions] for filename, instructions in global_state.plan] or [[""]],
            headers=["Proposed Plan"],
            interactive=True,
            col_count=(1, "static"),
            wrap=True
        )

    with gr.Row():
        with gr.Column(scale=8):
            msg = gr.Textbox(placeholder="Send a message to Sweep", label=None, elem_id="message_box")
        with gr.Column(scale=0.5):
            create_pr_button = gr.Button(value="Create PR", interactive=bool(global_state.plan))

    with gr.Row():
        tier_message = "⚡ Free Tier" if not user_info["is_paying_user"] else "💎 Pro Tier"
        remaining_tickets = user_info["remaining_tickets"]
        user_status = gr.Markdown(value=f"Logged in as {config.github_username} ({tier_message}), {remaining_tickets} tickets remaining.")

    def clear_inputs():
        global global_state
        global_state = State()
        config.state = global_state
        config.save()
        return [], [], [[""]]


    restart_button.click(clear_inputs, None, [file_names, chatbot, plan])

    file_names.change(get_files_update, repo_full_name, chatbot)

    searched = False
    selected_snippets = []
    file_to_str = {}


    def repo_name_change(repo_full_name):
        global installation_id
        try:
            try:
                installation_id = get_installation_id(repo_full_name)
            except Exception as e:
                raise gr.Error(str(e))
            assert installation_id
            config.installation_id = installation_id
            api_client.config = config
            config.save()
            return ""
        except Exception as e:
            config.repo_full_name = None
            config.installation_id = None
            config.save()
            api_client.config = config
            raise e


    def build_string():
        global selected_snippets
        global file_to_str
        for snippet in selected_snippets:
            file_name = snippet.file_path
            if file_name not in file_to_str:
                add_file_to_dict(file_name)
        snippets_text = "### Relevant snippets:\n" + "\n\n".join(
            [file_to_str[snippet.file_path] for snippet in selected_snippets])
        return snippets_text


    repo_full_name.change(repo_name_change, [repo_full_name], [msg])


    def add_file_to_dict(file_name):
        global file_to_str
        global path_to_contents
        global repo
        if file_name in path_to_contents:
            file_contents = path_to_contents[file_name]
        else:
            file_contents = repo.get_contents(file_name, ref=SweepConfig.get_branch(repo)).decoded_content.decode(
                'utf-8')
        file_contents_split = file_contents.split("\n")
        length = len(file_contents_split)
        backtick, escaped_backtick = "`", "\\`"
        preview = "\n".join(file_contents_split[:3]).replace(backtick, escaped_backtick)
        file_to_str[file_name] = f'{file_name}:0:{length}\n```\n{preview}\n...\n```'


    def file_names_change(file_names):
        global selected_snippets
        global file_to_str
        global path_to_contents
        selected_snippets = [
            Snippet(content=path_to_contents[file_name], start=0, end=path_to_contents[file_name].count('\n'),
                    file_path=file_name) for file_name in file_names]
        return file_names, build_string()


    file_names.change(file_names_change, [file_names], [file_names, snippets_text])

    def handle_message_submit(repo_full_name: str, user_message: str, history: list[tuple[str | None, str | None]]):
        if not repo_full_name:
            raise gr.Error("Set the repository name first")
        return gr.update(value="", interactive=False), history + [[user_message, None]], gr.Button.update(
            interactive=False)

    def _handle_message_stream(chat_history: list[tuple[str | None, str | None]], snippets_text, file_names, plan):
        global selected_snippets
        global searched
        if plan is None or plan == [[]] or plan == [[""]] or plan == [["", ""]]:
            plan = [["", ""]]
        message = chat_history[-1][0]
        yield chat_history, snippets_text, file_names, plan
        if not selected_snippets:
            searched = True
            # Searching for relevant snippets
            chat_history[-1][1] = "Searching for relevant snippets..."
            snippets_text = build_string()
            yield chat_history, snippets_text, file_names, plan
            logger.info("Fetching relevant snippets...")
            try:
                selected_snippets += api_client.search(chat_history[-1][0], 3)
            except Exception as e:
                raise gr.Error(str(e))
            snippets_text = build_string()
            file_names = [snippet.file_path for snippet in selected_snippets]
            yield chat_history, snippets_text, file_names, plan
            logger.info("Fetched relevant snippets.")
            chat_history[-1][1] = "Found relevant snippets."
            # Update using chat_history
            snippets_text = build_string()
            yield chat_history, snippets_text, file_names, plan

        # Generate response
        logger.info("...")
        chat_history.append([None, "..."])
        yield chat_history, snippets_text, file_names, plan
        chat_history[-1][1] = ""
        logger.info("Starting to generate response...")
        if len(chat_history) > 1 and "create pr" in message.lower():
            try:
                stream = api_client.stream_chat(
                    chat_history,
                    selected_snippets,
                    functions=[create_pr_function],
                    function_call=create_pr_function_call,
                )
            except Exception as e:
                raise gr.Error(str(e))
        else:
            stream = api_client.stream_chat(chat_history, selected_snippets)
        function_name = ""
        raw_arguments = ""
        raw_response = ""
        parsed_response = ""
        for chunk in stream:
            if chunk.get("content"):
                token = chunk["content"]
                raw_response += token
                parsed_response, plan = parse_response(raw_response)
                chat_history[-1][1] = parsed_response
                yield chat_history, snippets_text, file_names, plan
            if chunk.get("function_call"):
                function_call = chunk["function_call"]
                function_name = function_name or function_call.get("name")
                raw_arguments += function_call.get("arguments")
                chat_history[-1][1] = f"Calling function: `{function_name}`\n```json\n{raw_arguments}\n```"
                yield chat_history, snippets_text, file_names, plan
        if function_name:
            arguments = json.loads(raw_arguments)
            if function_name == "create_pr":
                assert "title" in arguments
                assert "summary" in arguments
                assert "plan" in arguments
                if "branch" not in arguments:
                    arguments["branch"] = arguments["title"].lower().replace(" ", "_").replace("-", "_")[:50]
                chat_history[-1][1] = pr_summary_template.format(
                    title=arguments["title"],
                    summary=arguments["summary"],
                    plan="\n".join([f"* `{item['file_path']}`: {item['instructions']}" for item in arguments["plan"]])
                )
                yield chat_history, snippets_text, file_names, plan
                plan = [(item["file_path"], item["instructions"]) for item in arguments["plan"]]
                yield chat_history, snippets_text, file_names, plan
            else:
                raise NotImplementedError


    def handle_message_stream(chat_history: list[tuple[str | None, str | None]], snippets_text, file_paths, plan):
        global global_state
        for chat_history, snippets_text, file_paths, plan in _handle_message_stream(chat_history, snippets_text,
                                                                                    file_paths, plan):
            if plan is None or plan == [[]] or plan == [[""]] or plan == [["", ""]]:
                plan = [["", ""]]
            if plan and isinstance(plan[0], list):
                plan = [item.split(":") for item in plan if ":" in item]
            global_state = State(
                chat_history=chat_history,
                snippets_text=snippets_text,
                file_paths=file_paths,
                plan=plan,
            )
            config.state = global_state
            config.save()
            if plan == [["", ""]]:
                yield chat_history, snippets_text, file_paths, [[""]]
            else:
                yield chat_history, snippets_text, file_paths, [(file_path + ": " + instructions,) for
                                                            file_path, instructions in plan]


    response = msg \
        .submit(handle_message_submit, [repo_full_name, msg, chatbot], [msg, chatbot, create_pr_button], queue=False) \
        .then(handle_message_stream, [chatbot, snippets_text, file_names, plan],
              [chatbot, snippets_text, file_names, plan]) \
        .then(lambda: [gr.update(interactive=True), gr.update(interactive=True)], None, [msg, create_pr_button], queue=False)


    def validate_branch_name(branch_name):
        # Replace any characters that are not alphanumeric or '-' or '_' with '_'
        valid_branch_name = re.sub('[^0-9a-zA-Z_-]', '_', branch_name)
        return valid_branch_name

    def on_create_pr_button_click(chat_history: list[tuple[str | None, str | None]], plan: list[tuple[str]]):
        title = chat_history[0][0]
        content = chat_history[-1][1]
        # Validate the branch name before it's used in the create_pr function
        valid_branch_name = validate_branch_name(title.lower().replace(" ", "_").replace("-", "_")[:50])
        chat_history.append((None, "⌛ Creating PR..."))
        yield chat_history, gr.Button.update(interactive=False)
        try:
            pull_request = api_client.create_pr(
                file_change_requests=[(item[:item.find(":")], item[item.find(":") + 1:]) for item, *_ in plan],
                pull_request={
                    "title": title,
                    "content": content,
                    "branch_name": valid_branch_name
                },
                messages=chat_history,
            )
            chat_history.append((None, f"✅ PR created at {pull_request['html_url']}"))
            webbrowser.open_new_tab(pull_request["html_url"])
            yield chat_history, gr.Button.update(interactive=True)
        except Exception as e:
            yield chat_history, gr.Button.update(interactive=True)
            raise gr.Error(str(e))


    create_pr_button.click(on_create_pr_button_click, [chatbot, plan], [chatbot, create_pr_button])

if __name__ == "__main__":
    demo.queue()
    demo.launch()



import typer

from sweepai.app.config import SweepChatConfig

epilog = "Sweep is a AI junior developer. Docs at https://docs.sweep.dev, install at https://github.com/apps/sweep-ai and support at https://discord.gg/sweep-ai."

typer_app = typer.Typer(epilog=epilog)


# @app.callback()
@typer_app.command()
def start():
    """
    Launch Sweep Chat in the browser
    """
    SweepChatConfig.load()
    from sweepai.app.ui import demo
    print("\033[93m⭐ Remember to star our repo at https://github.com/sweepai/sweep! \033[0m")
    demo.queue()
    demo.launch(inbrowser=True)


@typer_app.command()
def auth():
    """
    Reauthenticate with Github API for Sweep to work (for token expiry)
    """
    SweepChatConfig.load(recreate=True)
    print("Setup completed successfully!")
    print("\033[93m⭐ Remember to star our repo at https://github.com/sweepai/sweep! \033[0m")


def app():
    # hacky solution based on https://github.com/tiangolo/typer/issues/18#issuecomment-1577788949
    import sys
    commands = {'start', 'auth'}
    sys.argv.append('start') if sys.argv[-1] not in commands else None
    typer_app()


if __name__ == "__main__":
    app()
